import os
from unittest.mock import Mock, patch

from pkg_resources import parse_version
import pytest
from vr.common.models import ProcData
from vr.common.paths import get_container_path

from vr.runners import image


@pytest.fixture()
def runner():
    r = image.ImageRunner()
    r.config = ProcData({
        'app_name': 'myApp',
        'proc_name': 'proc.exe',
        'port': 1234,
        'release_hash': 'deadbeef',
        'version': '1.0',
        'config_name': 'config-name',
        'image_name': 'image-name',
        'cmd': 'command',
        # 'env': {},
    })
    r.ensure_image = Mock()
    r.ensure_container = Mock()
    r.ensure_char_devices = Mock()
    return r


class TestImageRunner(object):

    @patch('vr.runners.image.get_lxc_version')
    def test_setup(self, get_lxc_version_, runner):
        get_lxc_version_.return_value = parse_version('1.0.8')
        runner.setup()
        p = get_container_path(runner.config)
        assert os.path.exists(os.path.join(p, 'env.sh')), 'env.sh missing'
        assert os.path.exists(os.path.join(p, 'proc.sh')), 'proc.sh missing'
        assert os.path.exists(os.path.join(p, 'settings.yaml')), \
            'settings.yaml missing'
        assert os.path.exists(os.path.join(p, '../proc.lxc')), \
            'proc.lxc missing'
        assert os.path.isdir(os.path.join(p, '../work')), \
            'work_dir does not exists'

    @patch('vr.runners.image.get_lxc_version')
    def test_proc_lxc(self, get_lxc_version_, runner):
        get_lxc_version_.return_value = parse_version('1.0.8')
        runner.setup()
        p = get_container_path(runner.config)
        proc_lxc = open(os.path.join(p, '../proc.lxc'), 'r').read()

        assert 'lxc.network.type = none' in proc_lxc
        assert 'lxc.mount.entry = overlayfs ' in proc_lxc

    @patch('vr.runners.image.get_lxc_version')
    def test_proc_lxc_lxc_pre1(self, get_lxc_version_, runner):
        get_lxc_version_.return_value = parse_version('0.9.9')
        runner.setup()
        p = get_container_path(runner.config)
        proc_lxc = open(os.path.join(p, '../proc.lxc'), 'r').read()

        assert 'lxc.network.type = none' not in proc_lxc
        assert 'lxc.mount.entry = overlayfs ' in proc_lxc
        assert 'workdir' not in proc_lxc

    @patch('vr.runners.image.get_lxc_version')
    def test_proc_lxc_lxc_post2(self, get_lxc_version_, runner):
        get_lxc_version_.return_value = parse_version('2.0.1')
        runner.setup()
        p = get_container_path(runner.config)
        proc_lxc = open(os.path.join(p, '../proc.lxc'), 'r').read()

        assert 'lxc.network.type = none' in proc_lxc
        assert 'lxc.mount.entry = overlay ' in proc_lxc
        assert 'workdir' in proc_lxc



import shutil
import tempfile

import vr.common.paths


VR_ROOT = None


def pytest_configure():
    global VR_ROOT

    vr.common.paths.VR_ROOT = VR_ROOT = tempfile.mkdtemp()
    vr.common.paths.BUILDS_ROOT = VR_ROOT + '/builds'
    vr.common.paths.PROCS_ROOT = VR_ROOT + '/procs'
    vr.common.paths.RELEASES_ROOT = VR_ROOT + '/releases'
    vr.common.paths.IMAGES_ROOT = VR_ROOT + '/images'


def pytest_unconfigure():
    if VR_ROOT is not None:
        shutil.rmtree(VR_ROOT)



from __future__ import print_function, unicode_literals

import os
import stat

import path

from vr.common.paths import (
    get_container_path, get_lxc_work_path, VR_ROOT)
from vr.common.utils import (
    get_lxc_version, get_lxc_overlayfs_config_fmt,
    get_lxc_network_config)
from vr.runners.base import BaseRunner, mkdir, ensure_file, untar


IMAGES_ROOT = VR_ROOT + '/images'


def ensure_image(name, url, images_root, md5, untar_to=None):
    """Ensure OS image at url has been downloaded and (optionally) unpacked."""
    image_dir_path = os.path.join(images_root, name)
    mkdir(image_dir_path)
    image_file_path = os.path.join(image_dir_path, os.path.basename(url))
    ensure_file(url, image_file_path, md5)
    if untar_to:
        prepare_image(image_file_path, untar_to)


def prepare_image(tarpath, outfolder, **kwargs):
    """Unpack the OS image stored at tarpath to outfolder.

    Prepare the unpacked image for use as a VR base image.

    """
    outfolder = path.Path(outfolder)
    untar(tarpath, outfolder, **kwargs)

    # Some OSes have started making /etc/resolv.conf into a symlink to
    # /run/resolv.conf.  That prevents us from bind-mounting to that
    # location.  So delete that symlink, if it exists.
    resolv_path = outfolder / 'etc' / 'resolv.conf'
    if resolv_path.islink():
        resolv_path.remove().write_text('', encoding='ascii')


class ImageRunner(BaseRunner):
    """
    A runner that launches apps inside a container built around a whole OS
    image tarball.  Requires that the proc config contain keys for 'image_url'
    and 'image_name'.

    Image tarballs are stored in /apps/images/<image_name>/<filename>.

    Unpacked images are stored in /apps/images/<image_name>/contents
    """

    lxc_template_name = 'image.lxc'

    char_devices = (
        ('/dev/null', (1, 3), 0o666),
        ('/dev/zero', (1, 5), 0o666),
        ('/dev/random', (1, 8), 0o444),
        ('/dev/urandom', (1, 9), 0o444),
    )

    def setup(self):
        print("Setting up", self.container_name)
        mkdir(IMAGES_ROOT)
        self.ensure_image()
        self.make_proc_dirs()
        self.ensure_build()
        self.ensure_char_devices()
        self.write_proc_lxc()
        self.write_settings_yaml()
        self.write_proc_sh()
        self.write_env_sh()
        self.ensure_container()

    def ensure_image(self):
        """
        Ensure that config.image_url has been downloaded and unpacked.
        """
        image_folder = self.get_image_folder()
        if os.path.exists(image_folder):
            print(
                'OS image directory {} exists...not overwriting' .format(
                    image_folder))
            return

        ensure_image(
            self.config.image_name,
            self.config.image_url,
            IMAGES_ROOT,
            getattr(self.config, 'image_md5', None),
            self.get_image_folder()
        )

    def get_image_folder(self):
        return os.path.join(IMAGES_ROOT, self.config.image_name, 'contents')

    def get_proc_lxc_tmpl_ctx(self):
        proc_path = get_container_path(self.config)
        work_path = get_lxc_work_path(self.config)
        ctx = {
            'proc_path': proc_path,
            'image_path': self.get_image_folder(),
            'work_path': work_path,
            'network_config': get_lxc_network_config(get_lxc_version()),
            'memory_limits': self.get_lxc_memory_limits(),
            'volumes': self.get_lxc_volume_str(),
        }
        overlay_config_fmt = get_lxc_overlayfs_config_fmt(get_lxc_version())
        ctx['overlay_config'] = overlay_config_fmt % ctx
        return ctx

    def ensure_char_devices(self):
        for path_, devnums, perms in self.char_devices:
            fullpath = get_container_path(self.config) + path_
            ensure_char_device(fullpath, devnums, perms)


def ensure_char_device(path, devnums, perms):
    # Python uses the OS mknod(2) implementation which modifies the mode based
    # on the umask of the running process (at least on some Linuxes that were
    # tested).  Set this to 0 to make mknod apply the perms you actually
    # specify
    print("Making device nodes")
    if not os.path.exists(path):
        with tmp_umask(0):
            print(
                "mknod -m %o %s c %s %s"
                % (perms, path, devnums[0], devnums[1]))
            mkdir(os.path.dirname(path))
            mode = (stat.S_IFCHR | perms)
            os.mknod(path, mode, os.makedev(*devnums))
    else:
        print("%s already exists.  Skipping" % path)


class tmp_umask(object):
    """Context manager for temporarily setting the process umask"""
    def __init__(self, tmp_mask):
        self.tmp_mask = tmp_mask

    def __enter__(self):
        self.orig_mask = os.umask(self.tmp_mask)

    def __exit__(self, type, value, traceback):
        os.umask(self.orig_mask)


if __name__ == '__main__':
    ImageRunner.invoke()



#!/usr/bin/env python

# Project skeleton maintained at https://github.com/jaraco/skeleton

import io

import setuptools

with io.open('README.rst', encoding='utf-8') as readme:
    long_description = readme.read()

name = 'vr.runners'
description = 'Command line tools to launch procs.'
nspkg_technique = 'managed'
"""
Does this package use "native" namespace packages or
pkg_resources "managed" namespace packages?
"""

params = dict(
    name=name,
    use_scm_version=True,
    author='Brent Tubbs',
    author_email='brent.tubbs@gmail.com',
    description=description or name,
    long_description=long_description,
    url="https://github.com/yougov/" + name,
    packages=setuptools.find_packages(),
    include_package_data=True,
    namespace_packages=(
        name.split('.')[:-1] if nspkg_technique == 'managed'
        else []
    ),
    python_requires='>=2.7',
    install_requires=[
        'vr.common>=4.9.0',
        'requests>=1.2.0',
        'path.py',
    ],
    extras_require={
        'testing': [
            # upstream
            'pytest>=3.5',
            'pytest-sugar>=0.9.1',
            'collective.checkdocs',
            'pytest-flake8',

            # local
            'backports.unittest_mock',
        ],
        'docs': [
            # upstream
            'sphinx',
            'jaraco.packaging>=3.2',
            'rst.linker>=1.9',

            # local
        ],
    },
    setup_requires=[
        'setuptools_scm>=1.15.0',
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
    ],
    entry_points={
        'console_scripts': [
            'vrun = vr.runners.image:ImageRunner.invoke',
        ],
    },
)
if __name__ == '__main__':
    setuptools.setup(**params)



# pylint: disable=attribute-defined-outside-init,redefined-outer-name
from __future__ import print_function

import argparse
import hashlib
import os
import shutil
import stat
import tarfile

import pkg_resources
import requests
import yaml
import path
import six

from vr.common.paths import (
    get_container_name, get_buildfile_path, BUILDS_ROOT, get_app_path,
    get_container_path, get_proc_path, get_lxc_work_path)
from vr.common.models import ProcData
from vr.common.utils import (
    tmpdir, mkdir, lock_file, which, file_md5,
    get_lxc_version, get_lxc_network_config)


def get_version():
    try:
        return pkg_resources.get_distribution('vr.runners').version
    except Exception:
        return None


class BaseRunner(object):
    """
    >>> callable(BaseRunner.shell.lock)
    True
    """

    def main(self):
        self.commands = {
            'setup': self.setup,
            'run': self.run,
            'shell': self.shell,
            'uptest': self.uptest,
            'teardown': self.teardown,
        }

        # pylint: disable=unused-variable
        cmd_list = ', '.join(self.commands)
        parser = argparse.ArgumentParser()
        cmd_help = 'One of: {cmd_list}'.format(**locals())
        parser.add_argument('command', help=cmd_help)
        parser.add_argument('file', help="Path to proc.yaml file.")
        parser.add_argument(
            '--version', action='version', version=get_version())

        args = parser.parse_args()

        try:
            cmd = self.commands[args.command]
        except KeyError:
            msg = 'Command must be one of: {cmd_list}'.format(**locals())
            raise SystemExit(msg)

        with open(args.file, 'r+b') as fid:
            self.config = ProcData(yaml.safe_load(fid))

            # Lock the file for exclusive access. Some commands (such as shell
            # or uptest) may override the behavior by providing a 'lock'
            # attribute on the method.
            getattr(cmd, 'lock', lock_file)(fid)
            cmd()

    @property
    def container_name(self):
        return get_container_name(self.config)

    # pylint: disable=no-self-argument,no-member
    def __close_file(fid):
        fid.close()

    def setup(self):
        print("Setting up", self.container_name)
        self.make_proc_dirs()
        self.ensure_build()
        self.write_proc_lxc()
        self.write_settings_yaml()
        self.write_proc_sh()
        self.write_env_sh()

    def run(self):
        print("Running", self.container_name)
        self._lxc_start()

    def shell(self):
        print("Running shell for", self.container_name)
        self._lxc_start(special_cmd='/bin/bash')
    shell.lock = __close_file

    def untar(self):
        tarpath = get_buildfile_path(self.config)
        print("Untarring", tarpath)
        outfolder = get_app_path(self.config)
        owners = (self.config.user, self.config.group)
        untar(tarpath, outfolder, owners)

    def write_proc_sh(self):
        """
        Write the script that is the first thing called inside the
        container.  It sets env vars and then calls the real program.
        """
        print("Writing proc.sh")
        context = {
            'tmp': '/tmp',
            'home': '/app',
            'settings': '/settings.yaml',
            'envsh': '/env.sh',
            'port': self.config.port,
            'cmd': self.get_cmd(),
        }
        sh_path = os.path.join(get_container_path(self.config), 'proc.sh')
        rendered = get_template('proc.sh') % context
        with open(sh_path, 'w') as f:
            f.write(rendered)
        st = os.stat(sh_path)
        os.chmod(
            sh_path, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    def write_env_sh(self):
        print("Writing env.sh")
        envsh_path = os.path.join(get_container_path(self.config), 'env.sh')

        with open(envsh_path, 'w') as f:

            def _interpolate(val):
                if isinstance(val, six.string_types) and val.startswith('$'):
                    return os.environ.get(val[1:], '')
                return val

            def format_var(key, val):
                return 'export %s="%s"' % (key, _interpolate(val))

            e = self.config.env or {}
            env_str = '\n'.join(format_var(k, e[k]) for k in e) + '\n'
            f.write(env_str)

    def get_cmd(self):
        """
        If self.config.cmd is not None, return that.

        Otherwise, read the Procfile inside the build code, parse it
        (as yaml), and pull out the command for self.config.proc_name.
        """
        if self.config.cmd is not None:
            return self.config.cmd

        procfile_path = os.path.join(get_app_path(self.config), 'Procfile')
        with open(procfile_path, 'r') as f:
            procs = yaml.safe_load(f)
        return procs[self.config.proc_name]

    def ensure_container(self, name=None):
        """Make sure container exists. It's only needed on newer
        versions of LXC."""
        if get_lxc_version() < pkg_resources.parse_version('2.0.0'):
            # Nothing to do for old versions of LXC
            return

        if name is None:
            name = self.container_name

        args = [
            'lxc-create',
            '--name', name,
            '--template', 'none',
            '>', '/dev/null', '2>&1',
        ]
        os.system(' '.join(args))

    def ensure_build(self):
        """
        If self.config.build_url is set, ensure it's been downloaded to the
        builds folder.
        """
        if self.config.build_url:
            path = get_buildfile_path(self.config)

            # Ensure that builds_root has been created.
            mkdir(BUILDS_ROOT)
            build_md5 = getattr(self.config, 'build_md5', None)
            ensure_file(self.config.build_url, path, build_md5)

            # Now untar.
            self.untar()

    def write_settings_yaml(self):
        print("Writing settings.yaml")
        path = os.path.join(get_container_path(self.config), 'settings.yaml')
        with open(path, 'w') as f:
            f.write(
                yaml.safe_dump(self.config.settings, default_flow_style=False))

    def _lxc_start(self, special_cmd=None):
        args = self.get_lxc_args(special_cmd=special_cmd)
        os.execve(which('lxc-start')[0], args, {})

    def get_lxc_args(self, special_cmd=None):

        name = self.container_name
        if special_cmd:
            cmd = special_cmd
            # Container names must be unique, so to allow running a shell or
            # uptests next to the app container, add more
            # stuff to the name. Assumes that "special_cmd" are
            # ephemeral, so create a unique container name per
            # special_cmd, to allow creating too many containers (new
            # versions of LXC don't clean after themselves).
            # Use a hash of the command to eliminate characters that
            # are invalid filename characters.

            # Note: container name can't be too long, because of a
            # limitation of earlier versions of LXC
            uniq = hashlib.md5(special_cmd.encode('utf-8')).hexdigest()[:8]
            stuff = '-TMP' + uniq
            name += stuff

            self.ensure_container(name)
        else:
            cmd = 'run'

        # If we have an app_folder value, then we can use it to determine where
        # LXC debug logs can live (which is somewhere that build processes are
        # able to access if they want it).
        log_args = []
        if getattr(self.config, 'app_folder', None):
            log_args = [
                '--logpriority', 'debug',
                '--logfile', os.path.join(
                    self.config.app_folder, '.lxcdebug.log'),
            ]

        extra_params = []
        if get_lxc_version() >= pkg_resources.parse_version('2.0.0'):
            # Early versions of LXC either didn't have the
            # --foreground flag or defaulted to it.
            extra_params.append('--foreground')

        return [
            'lxc-start',
            '--name', name,
            '--rcfile', os.path.join(get_proc_path(self.config), 'proc.lxc'),
        ] + extra_params + log_args + [
            '--',

            # Note: using `su` seems to crash lxc container when
            # building certain Py3 projects.
            # See comments on:
            # https://bitbucket.org/yougov/vr.runners/commits/05ba82d54636db8e0e0791ee470742d0ef440715
            # Note: use short options (long options are not supported
            # by old version of sudo)

            'sudo',
            '-u', self.config.user,
            '-E',  # preserve environment
            '-s', '/bin/bash',  # shell
            '-c', 'cd /app;source /env.sh; exec /proc.sh "%s"' % cmd,
        ]

    def get_lxc_volume_str(self):
        content = ''
        # append lines to bind-mount volumes.
        volumes = getattr(self.config, 'volumes', []) or []
        volume_tmpl = "\nlxc.mount.entry = %s %s%s none bind 0 0"
        for outside, inside in volumes:
            content += volume_tmpl % (
                outside, get_container_path(self.config), inside)
        return content

    def uptest(self):
        # copy the uptester into the container. ensure it's executable.
        src = pkg_resources.resource_filename(
            'vr.runners', 'uptester/uptester')
        container_path = get_container_path(self.config)
        dest = os.path.join(container_path, 'uptester')
        shutil.copy(src, dest)
        st = os.stat(dest)
        os.chmod(dest, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        proc_name = getattr(self.config, 'proc_name', None)
        if proc_name:
            app_path = get_app_path(self.config)
            uptests_path = os.path.join(app_path, 'uptests', proc_name)
            if os.path.isdir(uptests_path):
                # run an LXC container for the uptests.
                inside_path = os.path.join('/app/uptests', proc_name)
                cmd = '/uptester %s %s %s ' % (inside_path, self.config.host,
                                               self.config.port)
                self._lxc_start(special_cmd=cmd)
            else:
                # There are no uptests for this proc.  Output an empty
                # JSON list.
                print("[]")
    uptest.lock = __close_file

    def teardown(self):
        """
        Delete the proc path where everything has been put.
        The build will be cleaned up elsewhere.
        """
        proc_path = get_proc_path(self.config)
        if os.path.isdir(proc_path):
            shutil.rmtree(proc_path)

    def make_proc_dirs(self):
        print("Making directories")

        proc_path = get_proc_path(self.config)
        mkdir(proc_path)

        container_path = get_container_path(self.config)
        mkdir(container_path)

        work_path = get_lxc_work_path(self.config)
        mkdir(work_path)

        volumes = getattr(self.config, 'volumes', None) or []
        for _, inside in volumes:
            mkdir(os.path.join(container_path, inside.lstrip('/')))

    def get_lxc_memory_limits(self):
        lines = []
        mem_limit = getattr(self.config, 'mem_limit', None)
        if mem_limit:
            lines.append(
                'lxc.cgroup.memory.limit_in_bytes = {limit}'.format(
                    limit=mem_limit))
        memsw_limit = getattr(self.config, 'memsw_limit', None)
        if memsw_limit:
            lines.append(
                'lxc.cgroup.memory.memsw.limit_in_bytes = {limit}'.format(
                    limit=memsw_limit))

        return '\n'.join(lines)

    def get_proc_lxc_tmpl_ctx(self):
        return {
            'proc_path': get_container_path(self.config),
            'network_config': get_lxc_network_config(get_lxc_version()),
            'memory_limits': self.get_lxc_memory_limits(),
            'volumes': self.get_lxc_volume_str(),
        }

    def write_proc_lxc(self):
        print("Writing proc.lxc")
        proc_path = get_proc_path(self.config)
        tmpl = get_template(self.lxc_template_name)
        content = tmpl % self.get_proc_lxc_tmpl_ctx()
        filepath = os.path.join(proc_path, 'proc.lxc')
        with open(filepath, 'w') as f:
            f.write(content)

    @classmethod
    def invoke(cls):
        "Instantiate and invoke the runner"
        cls().main()


def untar(tarpath, outfolder, owners=None, overwrite=True, fixperms=True):
    """
    Unpack tarpath to outfolder.  Make a guess about the compression based on
    file extension (.gz or .bz2).

    The unpacking of the tarfile is done in a temp directory and moved into
    place atomically at the end (assuming /tmp is on the same filesystem as
    outfolder).

    If 'owners' is provided, it should be a tuple in the form
    (username, groupname), and the contents of the unpacked folder will be set
    with that owner and group.

    If outfolder already exists, and overwrite=True (the default), the existing
    outfolder will be deleted before the new one is put in place. If outfolder
    already exists and overwrite=False, IOError will be raised.
    """

    # We don't use fixperms at all
    _ignored = fixperms  # noqa

    # make a folder to untar to
    with tmpdir():
        _, _, ext = tarpath.rpartition('.')

        if ext not in ('gz', 'bz2'):
            raise ValueError('tarpath must point to a .gz or .bz2 file')

        tf = tarfile.open(tarpath, 'r:' + ext)
        try:
            os.mkdir('contents')
            tf.extractall('contents')
        finally:
            tf.close()

        if owners is not None:
            contents = path.Path('contents')
            for item in contents.walk():
                if item.isdir():
                    # chown user:group
                    item.chown(*owners)
                    # chmod ug+xr
                    item.chmod('ug+xr')
                if item.isfile() and not item.islink():
                    # chown user:group
                    item.chown(*owners)
                    # chmod ug+rw
                    item.chmod('ug+rw')

        if os.path.isdir(outfolder):
            if overwrite:
                shutil.rmtree(outfolder)
            else:
                raise IOError(
                    ('Cannot untar %s because %s already exists and '
                     'overwrite=False') % (tarfile, outfolder))
        shutil.move('contents', outfolder)


def ensure_file(url, path, md5sum=None):
    """
    If file is not already at 'path', then download from 'url' and put it
    there.

    If md5sum is provided, and 'path' exists, check that file matches the
    md5sum.  If not, re-download.
    """

    if not os.path.isfile(path) or (md5sum and md5sum != file_md5(path)):
        download_file(url, path)


def download_file(url, path):
    with tmpdir():
        print("Downloading %s" % url)
        base = os.path.basename(path)
        with open(base, 'wb') as f:
            resp = requests.get(url, stream=True)
            resp.raise_for_status()
            shutil.copyfileobj(resp.raw, f)
        shutil.move(base, path)


def get_template(name):
    """
    Look for 'name' in the vr.runners.templates folder.  Return its contents.
    """
    path = pkg_resources.resource_filename('vr.runners', 'templates/' + name)
    with open(path, 'r') as f:
        return f.read()
